"""
APT helper module
"""
import apt
import lsb_release


def has_updates():
    """
    Check for package upgrades

    :return: Whether package upgrades are available
    :rtype: bool
    """
    cache = apt.Cache()
    cache.update()
    cache.open(None)
    cache.upgrade(dist_upgrade=True)
    if cache.get_changes():
        return True
    return False


def get_distro():
    """
    Get distribution information from lsb_release

    :return: Distribution information
    :rtype: dict
    """
    return lsb_release.get_distro_information()



"""
Settings module
"""
import os
from pathlib import Path

from dotenv import load_dotenv

try:
    load_dotenv()
except EnvironmentError:
    # XXX: Workaround for readthedocs  # pylint: disable=fixme
    pass

ENV_PREFIX = 'SATNOGS_CONFIG_'

SATNOGS_SETUP_UPGRADE_SCRIPT = os.getenv(
    ENV_PREFIX + 'SATNOGS_SETUP_UPGRADE_SCRIPT', 'satnogs-upgrade'
)
SATNOGS_SETUP_STAMP_DIR = os.getenv(
    ENV_PREFIX + 'SATNOGS_SETUP_STAMP_DIR',
    str(Path.home().joinpath('.satnogs'))
)
SATNOGS_SETUP_BOOTSTRAP_STAMP = os.getenv(
    ENV_PREFIX + 'SATNOGS_SETUP_BOOTSTRAP_STAMP', '.bootstrapped'
)
SATNOGS_SETUP_INSTALL_STAMP = os.getenv(
    ENV_PREFIX + 'SATNOGS_SETUP_INSTALL_STAMP', '.installed'
)
CONFIG_FILE = os.getenv(
    ENV_PREFIX + 'CONFIG_FILE',
    str(Path.home().joinpath('.satnogs', 'config.yml'))
)
ANSIBLE_DIR = os.getenv(
    ENV_PREFIX + 'ANSIBLE_DIR',
    str(Path.home().joinpath('.satnogs', 'ansible'))
)
ANSIBLE_PLAYBOOK = os.getenv(ENV_PREFIX + 'ANSIBLE_PLAYBOOK', 'local.yml')
ANSIBLE_URL = os.getenv(
    ENV_PREFIX + 'ANSIBLE_URL', 'https://gitlab.com/librespacefoundation'
    '/satnogs/satnogs-client-ansible.git'
)
ANSIBLE_BRANCH = os.getenv(ENV_PREFIX + 'ANSIBLE_BRANCH')



"""
SatNOGS support info module
"""
import json
import platform
from datetime import datetime, timezone

import psutil

from satnogsconfig._version import get_versions
from satnogsconfig.helpers import apt

__version__ = get_versions()['version']

del get_versions


class Support():
    """
    Create support information to be used for reporting bugs
    """
    def __init__(self, config, satnogs_setup, ansible):
        """
        Class constructor
        """
        self._config = config
        self._satnogs_setup = satnogs_setup
        self._ansible = ansible

    @property
    def info(self):
        """
        Support information

        :return: Support information dictionary
        :rtype: dict
        """
        config = self._config.config.copy()
        redacted_keys = ['satnogs_api_token']
        for key in redacted_keys:
            if config.get(key) is not None:
                config[key] = '[redacted]'
        data = {
            "versions":
                {
                    "satnogs-client":
                        self._satnogs_setup.satnogs_client_version,
                    "satnogs-client-ansible":
                        self._satnogs_setup.satnogs_client_ansible_version,
                    "satnogs-flowgraphs":
                        self._satnogs_setup.satnogs_flowgraphs_version,
                    "gr-satnogs": self._satnogs_setup.gr_satnogs_version,
                    "gr-soapy": self._satnogs_setup.gr_soapy_version,
                    "gnuradio": self._satnogs_setup.gnuradio_version,
                    "satnogs-config": __version__,
                },
            "state":
                {
                    "is-applied": self._satnogs_setup.is_applied,
                    "pending-tags": None,
                },
            "system":
                {
                    "date": datetime.now(timezone.utc).isoformat(),
                    "distribution": apt.get_distro(),
                    "pending-updates": apt.has_updates(),
                    "platform": dict(platform.uname()._asdict()),
                    "memory": dict(psutil.virtual_memory()._asdict()),
                    "disk": dict(psutil.disk_usage('/')._asdict()),
                },
            "configuration": config,
        }

        tags = self._satnogs_setup.tags
        if tags:
            data['state']['pending-tags'] = list(tags)

        return data

    def dump(self, *args, **kwargs):
        """
        Dump support information

        :return: JSON dump of support information
        :rtype: str
        """
        return json.dumps(self.info, *args, **kwargs)



"""
Configuration module
"""
import logging
from pathlib import Path

import yaml

LOGGER = logging.getLogger(__name__)


class Config():
    """
    Manage configuration file

    :param filename: File path of configuration
    :type filename: str
    """
    def __init__(self, filename):
        """
        Class constructor
        """
        self._filename = filename
        try:
            config = self._load_config()
        except FileNotFoundError:
            config = None
        self.config = config

    def _load_config(self):
        """
        Load and parse YAML configuration

        :return: Configuration dictionary
        :rtype: dict or NoneType
        """
        with open(self._filename, 'r') as file:
            return yaml.safe_load(file)

    def dump_config(self, to_file=False):
        """
        Dump configuration in YAML format

        :param to_file: Dump to file
        :type to_file: bool, optional
        :return: YAML configuration
        :rtype: str
        """
        if to_file:
            Path(self._filename
                 ).parent.mkdir(mode=0o700, parents=True, exist_ok=True)
            with open(self._filename, 'w') as file:
                yaml_config = yaml.dump(
                    self.config, file, default_flow_style=False
                )
            self.config = self._load_config()
        else:
            yaml_config = yaml.dump(self.config, default_flow_style=False)

        return yaml_config

    def clear_config(self):
        """
        Clear configuration file
        """
        open(self._filename, 'w').close()
        self.config = self._load_config()

    def get_variable(self, variable):
        """
        Get variable value from configuration

        :param variable: Variable to get the value
        :type variable: str
        :return: Value of variable
        :rtype: str or bool or NoneType
        """
        if self.config:
            return self.config.get(variable)

        return None

    def set_variable(self, variable, value):
        """
        Set variable value in configuration

        :param variable: Variable to set the value
        :type variable: str
        :param value: Value of variable
        :type value: str or bool or NoneType
        """
        if not self.config:
            self.config = {}
        if value is not None:
            self.config[variable] = value
        else:
            self.config.pop(variable, None)
        self.dump_config(to_file=True)



"""
satnogs-setup module
"""
import os
import subprocess
import sys
from pathlib import Path

import satnogsconfig.settings as settings


def get_package_version(package_name):
    """
    Get installed version of the given package

    :return: Version of the given package
    :rtype: str
    """
    try:
        result = subprocess.run(
            "dpkg-query --show -f='${{Version}}' {}".format(package_name),
            shell=True,
            stdout=subprocess.PIPE,
            check=True
        )
        return result.stdout.decode('utf-8').strip() or 'unknown'
    except subprocess.CalledProcessError:
        return 'unknown'


class SatnogsSetup():
    """
    Interract with satnogs-setup
    """
    def __init__(self):
        """
        Class constructor
        """
        self._satnogs_stamp_dir = settings.SATNOGS_SETUP_STAMP_DIR
        self._satnogs_upgrade_script = settings.SATNOGS_SETUP_UPGRADE_SCRIPT

    @staticmethod
    def restart(boot=False):
        """
        Restart satnogs-setup script

        :param boot: Whether to bootstrap or not
        :type boot: bool, optional
        """
        if boot:
            os.execlp('satnogs-setup', 'satnogs-setup', '-b')
        else:
            os.execlp('satnogs-setup', 'satnogs-setup')

    @property
    def is_applied(self):
        """
        Check whether configuration has been applied

        :return: Whether configuration has been applied
        :rtype: bool
        """
        install_stamp_path = Path(self._satnogs_stamp_dir).joinpath(
            settings.SATNOGS_SETUP_INSTALL_STAMP
        )
        if install_stamp_path.exists():
            with install_stamp_path.open(mode='r') as file:
                contents = file.read()
                if contents:
                    return False
            return True
        return False

    @is_applied.setter
    def is_applied(self, install):
        """
        Mark that configuration has been applied

        :param install: Configuration has been installed
        :type install: bool
        """
        install_stamp_path = Path(self._satnogs_stamp_dir).joinpath(
            settings.SATNOGS_SETUP_INSTALL_STAMP
        )
        if install:
            install_stamp_path.open(mode='w').close()
        else:
            try:
                install_stamp_path.unlink()
            except FileNotFoundError:
                pass

    @property
    def tags(self):
        """
        Get satnogs-setup tags

        :return: Set of tags
        :rtype: set
        """
        tags_path = Path(self._satnogs_stamp_dir
                         ).joinpath(settings.SATNOGS_SETUP_INSTALL_STAMP)
        if tags_path.exists():
            with tags_path.open(mode='r') as file:
                contents = file.read()
                if contents:
                    return set(contents.split(','))
        return None

    @tags.setter
    def tags(self, tags):
        """
        Set satnogs-setup tags

        :param tags: List of tags
        :type tags: list
        """
        if self.tags:
            new_tags = self.tags.copy()
        else:
            new_tags = set()
        new_tags.update(tags)
        tags_path = Path(self._satnogs_stamp_dir
                         ).joinpath(settings.SATNOGS_SETUP_INSTALL_STAMP)
        if tags_path.exists():
            with tags_path.open(mode='w') as file:
                file.write(','.join(new_tags))

    def upgrade_system(self):
        """
        Upgrade system packages
        """
        try:
            subprocess.run(
                self._satnogs_upgrade_script, shell=True, check=True
            )
        except subprocess.CalledProcessError:
            sys.exit(1)

    @property
    def satnogs_client_ansible_version(self):
        """
        Get installed SatNOGS Client Ansible version

        :return: Version of SatNOGS Client Ansible
        :rtype: str
        """
        try:
            result = subprocess.run(
                'cd "$HOME/.satnogs/ansible"'
                '&& TZ=UTC '
                'git show -s --format=%cd --date="format-local:%Y%m%d%H%M"',
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True
            )
            return result.stdout.decode('utf-8').strip() or 'unknown'
        except subprocess.CalledProcessError:
            return 'unknown'

    @property
    def satnogs_client_version(self):
        """
        Get installed SatNOGS Client version

        :return: Version of SatNOGS Client
        :rtype: str
        """
        try:
            result = subprocess.run(
                "/var/lib/satnogs/bin/pip show satnogs-client 2>/dev/null"
                " | awk '/^Version: / { print $2 }'",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True
            )
            return result.stdout.decode('utf-8').strip() or 'unknown'
        except subprocess.CalledProcessError:
            return 'unknown'

    @property
    def gr_satnogs_version(self):
        """
        Get installed gr-satnogs version

        :return: Version of gr-satnogs
        :rtype: str
        """
        return get_package_version('gr-satnogs')

    @property
    def satnogs_flowgraphs_version(self):
        """
        Get installed satnogs-flowgraphs version

        :return: Version of satnogs-flowgraphs
        :rtype: str
        """
        return get_package_version('satnogs-flowgraphs')

    @property
    def gr_soapy_version(self):
        """
        Get installed gr-soapy version

        :return: Version of gr-soapy
        :rtype: str
        """
        return get_package_version('gr-soapy')

    @property
    def gnuradio_version(self):
        """
        Get installed gnuradio version

        :return: Version of gr-soapy
        :rtype: str
        """
        return get_package_version('gnuradio')



"""
gr-satnogs module
"""


class GrSatnogs():
    """
    Get and set gr-satnogs configuration
    """
    @property
    def gr_satnogs_version(self):
        """
        Get gr-satnogs version
        """
        raise NotImplementedError

    def probe_soapysdr(self):
        """
        Probe SoapySDR hardware
        """
        raise NotImplementedError



"""
Ansible helper module
"""
import subprocess


class Ansible():
    """
    Call Ansible playbooks
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, ansible_dir):
        """
        Class constructor
        """
        self._ansible_dir = ansible_dir

    def run(self, playbooks, tags=None, extra_args=None):
        """
        Run Ansible playbook

        :param playbooks: List of playbooks
        :type playbooks: list
        :param tags: List of tags
        :type tags: list, optional
        :param extra_args: List of extra arguments to pass to Ansible
        :type extra_args: list, optional
        :return: Whether Ansible execution succeeded
        :rtype: bool
        """
        args = []
        if extra_args:
            args += extra_args
        if tags:
            args += ['-t', ','.join(tags)]
        if playbooks:
            args += playbooks

        try:
            subprocess.run(
                ['ansible-playbook'] + args, cwd=self._ansible_dir, check=True
            )
            return True
        except subprocess.CalledProcessError:
            return False

    def pull(self, playbooks, url, branch=None, tags=None, extra_args=None):
        """
        Pull and run Ansible playbook

        :param playbooks: List of playbooks
        :type playbooks: list
        :param url: Git URL to pull playbooks
        :type url: str
        :param branch: Git branch to pull playbooks
        :type branch: str, optional
        :param tags: List of tags
        :type tags: list, optional
        :param extra_args: List of extra arguments to pass to Ansible
        :type extra_args: list, optional
        :return: Whether Ansible execution succeeded
        :rtype: bool
        """
        # pylint: disable=too-many-arguments
        args = []
        if extra_args:
            args += extra_args
        if tags:
            args += ['-t', ','.join(tags)]
        if playbooks:
            args += playbooks
        if branch:
            args += ['-C', branch]
        try:
            subprocess.run(
                ['ansible-pull', '-d', self._ansible_dir, '-U', url] + args,
                check=True
            )
            return True
        except subprocess.CalledProcessError:
            return False
